/*
 * DFA-based (deterministic finite state automaton) implementation of a parser 
 * for a text file with an email message defined by following grammar:
 * <message>    ::= <sender> <recipient>+ <carboncopy>* <subject> <body>
 * <sender>     ::= "from:" <address>
 * <recipient>  ::= "to:" <address>
 * <carboncopy> ::= "cc:" <address>
 * <subject>    ::= "subject:" <string>+
 * <body>       ::= "body:" <string>*
 * <address>    ::= <id>'@'<id>
 * <id>         ::= ['a'..'z','A'..'Z','0'..'9','.','-','_']+
 * <string>     ::= ['a'..'z','A'..'Z','0'..'9',''..'','.',',','_']+
 * 
 * (assuming that <id> and <string> do not contain key tokens)
 */

import java.util.*;
import java.io.*;

public class DFAMailer {
	private Message msg;
	// Integer constants for states and labels of automaton
	private final int Init = 0, Sender = 1, SAddress = 2, Recipient = 3, RAddress = 4,
					  Copy = 5, CAddress = 6, Subject = 7, STitle = 8, Body = 9, Error = -1;
	private final int FROM = 0, TO = 1, CC = 2, SUBJECT = 3, BODY = 4, ADDRESS = 5, STRING = 6, ERROR = -1;
	// Tabular representation of automaton
	private final int NumStates = 10, NumLabels = 7;
	private int[][] automaton;
	private int currentState;
	
	public DFAMailer() {
		// Initialize automaton with default state
		automaton = new int [NumStates][NumLabels];
		for (int i = 0; i < NumStates; i++)
			for (int j = 0; j < NumLabels; j++)
				automaton[i][j] = Error;
		// Set appropriate transitions
		automaton[Init]     [FROM]    = Sender;
		automaton[Sender]   [ADDRESS] = SAddress;
		automaton[SAddress] [TO]      = Recipient;
		automaton[Recipient][ADDRESS] = RAddress;
		automaton[RAddress] [TO]      = Recipient;
		automaton[RAddress] [CC]      = Copy;
		automaton[RAddress] [SUBJECT] = Subject;
		automaton[Copy]     [ADDRESS] = CAddress;
		automaton[CAddress] [CC]      = Copy;
		automaton[CAddress] [SUBJECT] = Subject;
		automaton[Subject]  [STRING]  = STitle;
		automaton[STitle]   [STRING]  = STitle;
		automaton[STitle]   [BODY]    = Body;
		automaton[Body]     [STRING]  = Body;
	}
		
	public static void main(String[] args) {
		if (args.length != 1) {
			System.out.println("Usage: java DFAMailer <message file>");
			System.exit(0);
		}
		DFAMailer mailer = new DFAMailer();
		mailer.processMessage(args[0]);
	}

	private void processMessage(String fileName) {
		String buffer = new String();
		try {
			BufferedReader reader = new BufferedReader(new FileReader(fileName));
			while (reader.ready()) {
				buffer += reader.readLine() + " ";	
			}
			reader.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
			System.exit(1);
		}
		parseMessage(buffer);
		System.out.println("*** Message ***\n" + msg);
	}
	
	private void parseMessage(String str) {
		String titleString = new String();
		String bodyString = new String();
		msg = new Message();
		// Set initial state and iterate over all tokens
		currentState = Init;
		StringTokenizer tokenizer = new StringTokenizer(str);
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			int label = encodeToken(token);
			// Check for invalid token and inform user about expected labels in current state
			if (automaton[currentState][label] == Error) {
				System.out.print("-> Error in state \"" + decodeState(currentState) + "\" with token \"" + token + "\", expected one of:");
				for (int j = 0; j < NumLabels; j++)
					if (automaton[currentState][j] != Error) System.out.print(" \"" + decodeLabel(j) + "\"");
				System.out.print("\n");
			} else {
				// Update content of message according to current state and label
				if (label == ADDRESS) {
					switch (currentState) {
						case Sender:    msg.setSender(token); break;
						case Recipient: msg.getRecipients().add(token); break;
						case Copy:      msg.getCarbonCopies().add(token); break;
					}
				}
				if (label == STRING) {
					switch (currentState) {
						case Subject:
						case STitle:  titleString += token + " "; break;
						case Body:    bodyString += token + " "; break;
					}
				}
				if (label == BODY) {
					msg.setSubject(titleString);
				}
				// Perform transition to next state
				currentState = automaton[currentState][label];
			}
		}
		// Check for correct terminal state and update message body accordingly
		if (currentState == Body)
			msg.setBody(bodyString);
		else
			System.out.println("-> Error: Parsing stopped in non-final state \"" + decodeState(currentState) + "\"");
	}
	
	// Utility method for encoding of a token as corresponding integer constant
	private int encodeToken(String token) {
		if (token.equals("from:")) return FROM;
		if (token.equals("to:")) return TO;
		if (token.equals("cc:")) return CC;
		if (token.equals("subject:")) return SUBJECT;
		if (token.equals("body:")) return BODY;
		if (isAddress(token)) return ADDRESS;
		if (token.length() > 0) return STRING;
		return ERROR;
	}
	
	// Utility method for decoding of a state integer constant (used for error output)
	private String decodeState(int num) {
		switch (num) {
			case 0: return "Init";
			case 1: return "Sender";
			case 2: return "SAddress";
			case 3: return "Recipient";
			case 4: return "RAddress";
			case 5: return "Copy";
			case 6: return "CAddress";
			case 7: return "Subject";
			case 8: return "STitle";
			case 9: return "Body";
			default: return "Error";
		}
	}
		
	// Utility method for decoding of a label integer constant (used for error output)
	private String decodeLabel(int num) {
		switch (num) {
			case 0: return "from:";
			case 1: return "to:";
			case 2: return "cc:";
			case 3: return "subject:";
			case 4: return "body:";
			case 5: return "<address>";
			case 6: return "<string>";
			default: return "error";
		}
	}
		
	// Simple check for a valid email address
	private boolean isAddress(String adr) {
		StringTokenizer tokenizer = new StringTokenizer(adr, "@");
		if (tokenizer.countTokens() != 2) {
			return false;
		}
		return true;
	}
}
